from __future__ import division
from fractions import Fraction
import math


class MusicError(Exception):
    pass


accidentals_table = {
    0  : [],
    1  : ["f#"],
    2  : ["f#", "c#"],
    3  : ["f#", "c#", "g#"],
    4  : ["f#", "c#", "g#", "d#"],
    5  : ["f#", "c#", "g#", "d#", "a#"],
    6  : ["f#", "c#", "g#", "d#", "a#", "e#"],
    7  : ["f#", "c#", "g#", "d#", "a#", "e#", "b#"],
    -1 : ["bb"],
    -2 : ["bb", "eb"],
    -3 : ["bb", "eb", "ab"],
    -4 : ["bb", "eb", "ab", "db"],
    -5 : ["bb", "eb", "ab", "db", "gb"],
    -6 : ["bb", "eb", "ab", "db", "gb", "cb"],
    -7 : ["bb", "eb", "ab", "db", "gb", "cb", "fb"],
    }


def power_two_series(n):
    """Return a list of the power of two numbers that summed will give n.

    The power of two series is :math:`\\sum_{i=0}^{n-1} 2^i`.
    
    >>> power_two_series(7)
    [1, 2, 4]
    """

    number_elements = int(math.log(n + 1, 2))
    return [2**x for x in range(0, number_elements)]


def calculate_duration(dur, dots=0):
    """Calculate the total duration of a note with augmentation dots.

    :param dur: reciprocal of usual note values (e.g. 8th note has the value of 8)
    :param dots: number of duration dots
   
    The total duration of a note is :math:`\\sum_{i=0} \\frac{R}{2^i}`
    where R is the reciprocal of ``dur``. For instance, a
    note with a duration like '16..' will have a total duration of
    1/16 + 1/32 + 1/64 = 7/64.
    
    >>> calculate_duration(16, 2)
    Fraction(7, 64)
    """

    if dur == 0 or dur == "breve" or dur == "brevis":
        base = Fraction(2, 1)
    elif dur == "longa":
        base = Fraction(4, 1)
    elif dur == "maxima":
        base = Fraction(8, 1)
    elif isinstance(dur, int):
        base = Fraction().from_decimal(dur) ** -1
    else:
        raise MusicError, "I don't recognize duration {0}".format(dur)
    
    return sum([base / (2 ** x) for x in range(0, dots + 1)])


def accidental(acc):
    if acc:
        op = 1 if acc[0] == "#" else -1
    else:
        op = 0

    return len(acc) * op


def string_to_code(notename, acc, code):
    """
    >>> string_to_code('b', '#', 'base12')
    0
    """
    
    notes = "c d e f g a b".split()
    dic = {'base40': ([3, 9, 15, 20, 26, 32, 38], 40),
           'base12': ([0, 2, 4, 5, 7, 8, 11], 12)}

    code_list = dic[code][0]
    n = code_list[notes.index(notename.lower())]
    return  (n + accidental(acc)) % dic[code][1]


def frac_to_dur(n):
    """Return a string representation for a duration represented as a fraction.

    >>> frac_to_dur(Fraction(3, 8))
    '4.'

    This function will fail with more complicated rythms than dotted notes.
    """

    num = n.numerator
    den = n.denominator
    
    if num == 1:
        return str(den)
    elif num in [3, 7, 15]:
        s = power_two_series(num)
        dur = int(den/s[-1])
        dot = "." * int(math.log(num + 1, 2) - 1)
        return "{0}{1}".format(dur, dot)
    else:
        raise MusicError, "I don't know how to work with duration " + str(n)


def notename_to_humdrum(notename, octave):
    """
    >>> note_to_humdrum('Cbb', 5)
    cc--
    """
    
    note = notename[0]
    acc = notename[1:].replace("b", "-")

    if octave > 3:
        return (note.lower() * (octave - 3)) + acc
    else:
        return (note.upper() * (4 - octave)) + acc


def notename_to_lily(notename, octave):
    """
    >>> note_to_lily('Cb', 5)
    cb''
    """

    note = notename[0]
    acc = notename[1:].replace("b", "es").replace("#", "is")

    if octave > 3:
        o = "'" * (octave - 3)
    else:
        o = "," * (3 - octave)

    return note.lower() + acc + o



import music


class Score(list):
    def get_spine(self, n):
        tmp = []
        for item in self:
            if type(item) is list:
                tmp.append(item[n])
            else:
                tmp.append(item)
        return tmp

    def get_spine_simple(self, n):
        return [x[n] for x in self if type(x) is list]

    def __init__(self, *args):
        list.__init__(self, args)
        self.title = ""
        self.composer = ""
        self.filename = ""
        self.split_spine = False
        self.join_spine = False
        self.spine_types = []
        self.spine_names = []
        self.measure_numbers = []


class Record(object):
    def __init__(self, name, data):
        self.name = name
        self.data = data


class Comment(object):
    def __init__(self, data, level):
        self.data = data
        self.level = level


class Tandem(object):
    def __init__(self, type, data):
        self.type = type
        self.data = data


class Exclusive(object):
    def __init__(self, name):
        self.name = name


class Note(object):
    def __init__(self, name, dur=None):
        self.name = name
        self.duration = dur
        self.articulations = []
        self.beams = []
        self.octave = 4
        self.code = None
        self.system = None
        self.type = None


class MultipleStop(list):
    pass


class Bar(object):
    def __init__(self, number, repeat_begin=False,
                 repeat_end=False, double=False):
        self.number = number or ""
        self.repeat_begin = repeat_begin
        self.repeat_end = repeat_end
        self.double = double


class Rest(object):
    def __init__(self, dur, wholenote=False):
        self.duration = dur
        self.print_as_whole = wholenote


class SpinePath(object):
    def __init__(self, type):
        self.type = type


class BlankLine(object):
    pass


class NullToken(object):
    pass


class Dynam(object):
    def __init__(self, data):
        self.data = data


class UnknownType(str):
    pass

    
def make_notes(notes):
    """return a :class:`score.Score` with the notes written as a string."""

    lst = [Note(n) for n in notes.split()]
    return Score(*lst)



from __future__ import print_function
from __future__ import division
from collections import defaultdict
from itertools import izip, count
import re
import utils
import music
from score import (Score, Record, Comment, Tandem, Exclusive, UnknownType,
                   Note, MultipleStop, Bar, Rest, NullToken, BlankLine,
                   SpinePath)


debug = False


class KernError(Exception):
    """Exception class for kern data."""
    pass


def kern_error(message):
    """Helper function to raise parsing errors."""

    raise KernError(message)


def replace_flats(string):
    return string.replace("-", "b")


## Parse kern

art = {'n': "natural",
       '/': "up-stem",
       '\\': "down-stem",
       'o': "harmonic",
       't': "trill-st",
       'T': "trill-wt",
       'S': "turn",
       '$': "inverted-turn",
       'R': "end-with-turn",
       'u': "down-bow",
       'H': "glissando-start",
       'h': "glissando-end",
       ';': "fermata",
       'Q': "gruppetto",
       'p': "app-main-note",
       'U': "mute",
       '[': "tie-start",
       ']': "tie-end",
       '_': "tie-midle",
       '(': "slur-start",
       ')': "slur-end",
       '{': "phrase-start",
       '}': "phrase-end",
       '\'': "staccato",
       's': "spiccato",
       '\\': "pizzicato",
       '`': "staccatissimo",
       '~': "tenuto",
       '^': "accent",
       ':': "arpeggiation",
       'v': "up-bow",
       'z': "sforzando",
       ',': "breath",
       'm': "mordent-st",
       'w': "inverted-mordent-st",
       'M': "mordent-wt",
       'W': "inverted-mordent-wt"}

beams = {'L': 'beam-start',
         'J': 'beam-end',
         'K': 'beam-partial-right',
         'k': 'beam-partial-left'}


types = {'dur': ("0123456789", "Duration must be together."),
         'note': ("abcdefgABCDEFG", "Notes must be together."),
         'dot': (".", "Dots must be together or after a number."),
         'acc': ("#-", "Accs."),
         'art': (art,),
         'beam': (beams,),
         'rest': ("r", "Rest."),
         'acciac': ("q", "App"),
         'app': ("P",)}


def parse_kern_note(note, accs):
    """Return a string representation of a note from the tokenizer representation.

    Humdrum represents flats as dashes ('-'); we replace them with 'b's:
    
    >>> parse_kern_note('cc', '--')
    'Cbb'
    """

    return note[0].upper() + replace_flats("".join(accs))


def parse_kern_octave(note, accs):
    """Calculate the octave of a note in the kern representation.

    Since kern uses repetition and case to indicate octave, we need to
    parse it. The central octave is 4, these are the most common
    octaves: ccc = 6, cc = 5, c = 4, C = 3, CC = 2, CCC = 1, CCCC = 0

    This is a basic example of use:
    
    >>> parse_kern_octave('bb', '')
    5

    And we need to take care of the cases where an accidental will
    change the octave:
    
    >>> parse_kern_octave('bb', '#')
    6
    """
    
    size = len(note)

    if size > 0:
        if note[0].islower():
            octave = 3 + size
        else:
            octave = -size + 4
            assert size <= 4, "octave can't be lower than 0, the value is " + str(octave)
    else:
        kern_error("Note can't be empty.")

    n = music.string_to_code(note[0], "", "base12")
    a = music.accidental(accs)

    return octave + ((n + a) // 12)


def kern_tokenizer(item, linen=1):
    tokens = defaultdict(list)

    def _is(char, type):
        return char in types[type][0]

    def parse(char, key, cond):
        tokens[key].append(char) if cond else kern_error(types[key][0])

    for i in range(0, len(item)):
        p = '' if i == 0 else item[i - 1]
        c = item[i]

        if _is(c, 'dur'):
            parse(c, 'dur', (not p or not tokens['dur'] or _is(p, 'dur')))
        elif _is(c, 'note'):
            parse(c, 'note', (not p or not tokens['note'] or _is(p, 'note')))
        elif _is(c, 'dot'):
            parse(c, 'dot', (_is(p, 'dur') or _is(p, 'dot')))
        elif _is(c, 'acc'):
            parse(c, 'acc', _is(p, 'note') or (_is(p, 'acc') and c == p))
        elif _is(c, 'rest'):
            parse(c, 'rest', (_is(p, 'rest') or not tokens['rest']))
        elif _is(c, 'app'):
            parse(c, 'app', (tokens['note'] and tokens['dur']))
        elif _is(c, 'art'):
            tokens['art'].append(art[c])
        elif _is(c, 'beam'):
            tokens['beam'].append(beams[c])
        elif _is(c, 'acciac'):
            tokens['acciac'].append(c)
        else:
            if debug:
                print("Humdrum character not recognized: " + c)
    return tokens


def parse_kern_item(item, lineno=1, itemno=1):
    tokens = kern_tokenizer(item, lineno)

    if (not tokens['dur']) and ((not tokens['acciac']) or (not tokens['app'])):
        kern_error("Duration can't be NULL.")

    if (tokens['note'] and tokens['rest']):
        kern_error("A note can't have a pitch and a rest.")

    dur = int("".join(tokens['dur']))
    dots = len(tokens['dot'])
    duration = music.calculate_duration(dur, dots)
    
    if tokens['note']:
        notename = "".join(tokens['note'])
        acc = "".join(tokens['acc'])
        name = parse_kern_note(notename, acc)
        note = Note(name, duration)
        note.articulations = tokens['art']
        note.beams = tokens['beam']
        note.octave = parse_kern_octave(notename, acc)
        note.code = music.string_to_code(notename[0], acc, "base40")
        note.system = "base40"
        note.type = "kern"
        return note
    elif tokens['rest']:
        wholeNote = not bool(tokens['rest'] or len(tokens['rest']) >= 1)
        return Rest(duration, wholeNote)
    else:
        kern_error("Kern data must have a note or rest.")


def parse_kern(item, linen=1, itemno=1):
    s = item.split(" ")

    if not item:
        kern_error("Kern item shouldn't be empty.")
    elif len(s) == 1:
        return parse_kern_item(item, linen, itemno)
    else:
        return MultipleStop([parse_kern_item(i, linen, itemno) for i in s])


## parse elements


def parse_spine_path(item):
    d = {'*-': "spine-end",
         '*+': "spine-add",
         '*^': "spine-split",
         '*v': "spine-join",
         '*x': "spine-swap"}

    return SpinePath(d[item])


def parse_bar(item):
    """Search a string for bar elements and return a :class:`score.Bar`

    This function will search for the bar number, if a bar begins or
    ends a repetition and if it's a double bar. Humdrum has a bunch of
    syntax for visual bar lines that we don't parse. See :ref:`todo`.
    """
    
    return Bar(utils.search_string("[0-9]+([a-z]+)?", item),
               bool(utils.search_string(":\\||:!", item)),
               bool(utils.search_string("\\|:|!:", item)),
               bool(utils.search_string("==", item)))


def parse_clef(item):
    dic = {"F4": "bass",
           "G2": "treble",
           "C3": "alto",
           "C4": "tenor",
           "Gv2": "tenor8",
           "C1": "soprano",
           "C2": "mezzo",
           "X" : "perc",
           # we have these as defaults
           "C" : "alto",
           "F" : "bass",
           "G" : "treble"}

    return Tandem("clef", dic[item])


def parse_expansion_list(item):
    return Tandem("expansion-list", item[:-1].split(','))


def parse_key_signature(item):
    item = item[1:-1]
    split_item = [item[i] + item[i+1] for i in range(0, len(item), 2)]
    list_keys = [replace_flats(x.lower().strip()) for x in split_item]
    size = len(list_keys)
    sharps = music.accidentals_table[size]
    flats = music.accidentals_table[-size]

    if list_keys == sharps:
        return Tandem("key-signature", size)
    elif list_keys == flats:
        return Tandem("key-signature", -size)
    else:
        return Tandem("key-signature", list_keys)


def parse_key(item):
    return Tandem("key", replace_flats(item))


def parse_tandem(item):
    if item.startswith("*clef"):
        return parse_clef(item[5:])
    elif item.startswith("*IC"):
        return Tandem("instr-class", item[3:])
    elif item.startswith("*IG"):
        return Tandem("instr-group", item[3:])
    elif item.startswith("*ITr"):
        return Tandem("transposing", item[4:])
    elif item.startswith("*I:"):
        return Tandem("instrument", item[3:])
    elif item.startswith("*I"):
        return Tandem("instrument", item[2:])
    elif item.startswith("*k"):
        return parse_key_signature(item[2:])
    elif item.startswith("*MM"):
        return Tandem("tempo", float(item[3:]))
    elif item.startswith("*M"):
        return Tandem("meter", item[2:])
    elif item.startswith("*tb"):
        return Tandem("timebase", float(item[3:]))
    elif item.startswith("*>["):
        return parse_expansion_list(item[3:])
    elif item.startswith("*>"):
        return Tandem("label", item[2:])
    elif item.endswith(":"):
        return parse_key(item[1:-1])
    else:
        return Tandem(None, item[1:])


def parse_data(data_type, item, lineno=1, itemno=1):
    """Apply the right function to parse ``data_type``.

    For instance, if ``data_type`` is 'kern', then ``parse_data`` will
    apply ``parse_kern`` to the rest of arguments. ``parse_data`` has
    a dispatch table to match the data type to the function. When the
    parser for a new humdrum data type is implemented, the dispatch
    table need to be updated. If an unknown type is found the item
    will be returned without any parsing.
    """

    def unknown_type(item, lineno=1, itemno=1):
        return UnknownType(item)

    dispatch = {"kern": parse_kern}

    return dispatch.get(data_type, unknown_type)(item, lineno, itemno)


## basic parser


def parse_item(item, score, lineno=1, itemno=1):
    """Parse each item of a humdrum spine.

    We can parse general items like :class:`score.Bar`,
    :class:`score.Exclusive`, :class:`score.Tandem`,
    :class:`score.Comment`, and :class:`score.NullToken` without
    previous information since the type of data is defined in the
    string itself (e.g. the '**' in '**kern' defines it as an
    exclusive interpretation). However, we can't parse other data
    records without knowing their spine type. When an exclusive
    interpretation is found the spine type is appended to
    :attr:`Score.spine_types`. This attribute can be used to know the
    spine type of an item.
    """

    if type(item) is list:
        return [parse_item(i, score, lineno, itemno) for i in item]
    elif item.startswith("="):
        return parse_bar(item)
    elif item.startswith("**"):
        spine_type = item[2:]
        score.spine_types.append(spine_type)
        return Exclusive(spine_type)
    elif item in ["*-", "*+", "*^", "*v", "*x"]:
        return parse_spine_path(item)
    elif item.startswith("*"):
        if item.startswith("*I:"):
            score.spine_names.append(item[3:])
        elif (item.startswith("*I") and not item.startswith("*IC")
              and len(score.spine_names) != itemno - 1):
            score.spine_names.append(item[2:])

        return parse_tandem(item)
    elif item.startswith("!"):
        return parse_comment(item)
    elif item == ".":
        return NullToken()
    else:
        spine_type_list = score.spine_types
        if len(spine_type_list) == 0:
            kern_error("Can't parse an item without knowing the spine type.")
        else:
            data_type = score.spine_types[itemno]
        return parse_data(data_type, item, lineno, itemno)


def parse_reference_record(line):
    """Separate the data in the Reference from the ! character.

    Since a reference record must start with 3 exclamation marks we
    don't need to use regular expressions to separate the actual
    comment from the comment character. We split the line by the first
    occurrence of ':' in order to prevent splitting a colon inside the
    text, such as '!!!OTL: Title: more stuff'.
    """

    s = line.split(":", 1)
    return Record(s[0][3:].strip(), s[1].strip())


def parse_comment(line):
    """Separate the actual comment from the comment character."""

    match = re.match("^(!+)(.*)$", line)
    return Comment(match.group(2).strip(), len(match.group(1)))


def parse_line(line, score, lineno=1):
    """Parse a line, append the parsed result into the score and return the score

    A line can be a BlankLine, a reference record, a comment, or have
    tabular data (spines) in which case we parse each item
    individually with :func:`humdrum.parse_item`. Both reference
    record and comments start with !, but a reference record starts
    with 3 exclamation marks (!!!) while a comment starts with one,
    two, four or more exclamation marks. Global comments have more
    than one exclamation marks and will be catched by this function.
    Local comments have only one exclamation mark and are applied to an
    individual spine. A local comment will be catched in
    :func:`humdrum.parse_item()`.
    """

    if utils.search_string(r"^[ \t]*$", line) is not None:
        score.append(BlankLine())
    elif utils.search_string(r"^!{3}[a-zA-Z ]+", line):
        record = parse_reference_record(line)
        if record.name.startswith("COM"):
            score.composer = record.data
        elif record.name.startswith("OTL"):
            score.title = record.data

        score.append(record)
    elif utils.search_string(r"^(!{2})|(!{4,})[a-zA-Z ]+", line):
        score.append(parse_comment(line))
    else:
        s = line.split("\t")

        if score.split_spine:
            n = score.split_spine
            s[n:n+2] = [s[n:n+2]]

        if "*^" in s:
            score.split_spine = s.index("*^")

        for i in s:
            if type(i) is list:
                if "*v" in i:
                    score.split_spine = False
            elif "*v" == i:
                score.split_spine = False

        parsed = [parse_item(i, score, lineno, n) for i, n in izip(s, count())]
        score.append(parsed)
    return score


def parse_string(string):
    """Parse a string with humdrum data  and return a :class:`score.Score`.

    This function is useful mainly for tests and to parse data
    directly from user input. To parse large quantities of data you
    should use :func:`parse_file()` instead.
    """

    s = Score()
    for line, lineno in izip(string.split('\n'), count(1)):
        parse_line(line, s, lineno)
    return s


def parse_file(filename):
    """Parse a humdrum file and return a :class:`score.Score`."""

    with open(filename) as f:
        s = Score()
        for line, lineno in izip(f, count(1)):
            parse_line(line.rstrip(), s, lineno)
        return s


if __name__ == "__main__":
    #f = parse_file("../data/k160-02.krn")
    f = parse_file("../data/test.krn")
    #f = parse_file("../data/001.krn")
    #f = parse_file("../data/split.krn")
    #f = parse_file("../data/wtc-1/wtc1f03.krn")



import re
import sys
import itertools


def ispython3():
    """Return True if the main version of python is 3"""

    return sys.version[:1] == '3'


def search_string(pattern, string):
    """Like re.search but return the string that matches the pattern
    instead of a Match object."""

    tmp = re.search(pattern, string)
    if tmp:
        return tmp.group()


def flatten(listOfLists):
    "Flatten one level of nesting"
    return itertools.chain.from_iterable(listOfLists)



# from http://pypi.python.org/pypi/multimethod/0.2
# Author: Aric Coady <coady at bent-arrow com>

"""
Multiple argument dispacthing.

Call multimethod on a variable number of types.
It returns a decorator which finds the multimethod of the same name,
creating it if necessary, and adds that function to it.  For example:

    @multimethod(*types)
    def func(*args):
        ...

'func' is now a multimethod which will delegate to the above function,
when called with arguments of the specified types.  If an exact match
can't be found, the next closest method will be called (and cached).
A function can have more than one multimethod decorator.

See tests for more example usage.
"""

import sys
from itertools import imap, izip

class DispatchError(TypeError):
    pass

class signature(tuple):
    "A tuple of types that supports partial ordering."
    __slots__ = ()
    def __le__(self, other):
        return len(self) <= len(other) and all(imap(issubclass, other, self))
    def __lt__(self, other):
        return self != other and self <= other
    def __sub__(self, other):
        "Return relative distances, assuming self >= other."
        return [list(left.__mro__).index(right) for left, right in izip(self, other)]

class multimethod(dict):
    "A callable directed acyclic graph of methods."
    def __new__(cls, *types):
        "Return a decorator which will add the function."
        namespace = sys._getframe(1).f_locals
        def decorator(func):
            if isinstance(func, cls):
                self, func = func, func.last
            elif func.__name__ in namespace:
                self = namespace[func.__name__]
            else:
                self = dict.__new__(cls)
                self.__name__, self.cache = func.__name__, {}
            self[types] = self.last = func
            return self
        return decorator
    def parents(self, types):
        "Find immediate parents of potential key."
        parents, ancestors = set(), set()
        for key, (value, superkeys) in self.items():
            if key < types:
                parents.add(key)
                ancestors |= superkeys
        return parents - ancestors
    def __getitem__(self, types):
        return dict.__getitem__(self, types)[0]
    def __setitem__(self, types, func):
        self.cache.clear()
        types = signature(types)
        parents = self.parents(types)
        for key, (value, superkeys) in self.items():
            if types < key and (not parents or parents & superkeys):
                superkeys -= parents
                superkeys.add(types)
        dict.__setitem__(self, types, (func, parents))
    def __delitem__(self, types):
        self.cache.clear()
        dict.__delitem__(self, types)
        for key, (value, superkeys) in self.items():
            if types in superkeys:
                dict.__setitem__(self, key, (value, self.parents(key)))
    def super(self, *types):
        "Return the next applicable method of given types."
        types = signature(types)
        keys = self.parents(types)
        if keys:
            return self[min(keys, key=types.__sub__)]
        raise DispatchError("%s%s: no methods found" % (self.__name__, types))
    def __call__(self, *args, **kwargs):
        "Resolve and dispatch to best method."
        types = tuple(imap(type, args))
        try:
            func = self.cache[types]
        except KeyError:
            func = self.cache[types] = self[types] if types in self else self.super(*types)
        return func(*args, **kwargs)

class strict_multimethod(multimethod):
    "A multimethod which requires a single unambiguous best match."
    def super(self, *types):
        "Return the next applicable method of given types."
        keys = self.parents(signature(types))
        if len(keys) == 1:
            return self[keys.pop()]
        raise DispatchError("%s%s: %d methods found" % (self.__name__, types, len(keys)))



from __future__ import print_function
from score import (Score, Record, Comment, Tandem, Exclusive, UnknownType,
                   Note, MultipleStop, Bar, Rest, NullToken, BlankLine,
                   SpinePath)
from multimethod import multimethod
import sys
import music


humdrum_table = {
    'natural': "n",
    "up-stem": '/',
    "down-stem": '\\',
    "harmonic": 'o',
    "trill-st": 't',
    "trill-wt": 'T',
    "turn": 'S',
    "inverted-turn": '$',
    "end-with-turn": 'R',
    "down-bow": 'u',
    "glissando-start": 'H',
    "glissando-end": 'h',
    "fermata": ';',
    "gruppetto": 'Q',
    "app-main-note": 'p',
    "mute": 'U',
    "tie-start": '[',
    "tie-end": ']',
    "tie-midle": '_',
    "slur-start": '(',
    "slur-end": ')',
    "phrase-start": '{',
    "phrase-end": '}',
    "staccato": '\'',
    "spiccato": 's',
    "pizzicato": '\\',
    "staccatissimo": '`',
    "tenuto": '~',
    "accent": '^',
    "arpeggiation": ':',
    "up-bow": 'v',
    "sforzando": 'z',
    "breath": ',',
    "mordent-st": 'm',
    "inverted-mordent-st": 'w',
    "mordent-wt": 'M',
    "inverted-mordent-wt": 'W',
    'beam-start': 'L',
    'beam-end': 'J',
    'beam-partial-right': 'K',
    'beam-partial-left': 'k'
    }


class ConvertError(Exception):
    pass


## humdrum
@multimethod(Score)
def show_as_humdrum(self):
    for item in self:
        if type(item) is list:
            print("\t".join([show_as_humdrum(x) for x in item]))
        else:
            print(show_as_humdrum(item))


@multimethod(Record)
def show_as_humdrum(self):
    return "!!! {0}: {1}".format(self.name, self.data)


@multimethod(Comment)
def show_as_humdrum(self):
    return "{0} {1}".format(self.level * "!", self.data)


@multimethod(Tandem)
def show_as_humdrum(self):
    return "*{0}{1}".format(self.type, self.data)


@multimethod(Exclusive)
def show_as_humdrum(self):
    return "**{0}".format(self.name)


@multimethod(Note)
def show_as_humdrum(self):
    name = music.notename_to_humdrum(self.name, self.octave)
    dur = music.frac_to_dur(self.duration)
    return "{1}{0}".format(name, dur)


@multimethod(MultipleStop)
def show_as_humdrum(self):
    return " ".join([show_as_humdrum(x) for x in self])


@multimethod(list)
def show_as_humdrum(self):
    return "\t".join([show_as_humdrum(x) for x in self])


@multimethod(Bar)
def show_as_humdrum(self):
    return "={0}".format(self.number)


@multimethod(SpinePath)
def show_as_humdrum(self):
    return "*{0}".format(self.type)


@multimethod(Rest)
def show_as_humdrum(self):
    dur = music.frac_to_dur(self.duration)
    return "{0}r".format(dur)


@multimethod(BlankLine)
def show_as_humdrum(self):
    return "\n"


@multimethod(NullToken)
def show_as_humdrum(self):
    return "."


@multimethod(UnknownType)
def show_as_humdrum(self):
    return self


## lily

def print_as_lily(sco, afile):
    with open(afile, 'w') as f:
        show_as_lily(sco, f)


@multimethod(Score)
def show_as_lily(self, stream=sys.stdout):
    spines = [s for s in self.spine_types if s == "kern"]

    print("""\\header {{
    title = \"{0}\"
    composer = \"{1}\"
    }}
    """.format(self.title, self.composer), file=stream)

    print("{\n<<\n", file=stream)
    for spine in reversed(range(0, len(spines))):
        print("\\new Staff {", file=stream)
        print(" ".join([show_as_lily(i) for i in self.get_spine_simple(spine)]), file=stream)
        print("}", file=stream)
    print(">>\n}\n", file=stream)


@multimethod(Note)
def show_as_lily(self):
    name = music.notename_to_lily(self.name, self.octave)
    dur = music.frac_to_dur(self.duration)
    return "{0}{1}".format(name,  dur)


@multimethod(Rest)
def show_as_lily(self):
    dur = music.frac_to_dur(self.duration)
    return "r{0}".format(dur)


@multimethod(MultipleStop)
def show_as_lily(self):
    return " ".join([show_as_lily(x) for x in self])


@multimethod(Bar)
def show_as_lily(self):
    return "|\n"


@multimethod()
def show_as_lily(self):
    return ""
